#ifndef KQEMU_H
#define KQEMU_H

#define KQEMU_VERSION 0x010100

struct kqemu_segment_cache {
    uint32_t selector;
    unsigned long base;
    uint32_t limit;
    uint32_t flags;
};

struct kqemu_cpu_state {
#ifdef __x86_64__
    unsigned long regs[16];
#else
    unsigned long regs[8];
#endif
    unsigned long eip;
    unsigned long eflags;

    uint32_t dummy0, dummy1, dumm2, dummy3, dummy4;

    struct kqemu_segment_cache segs[6]; /* selector values */
    struct kqemu_segment_cache ldt;
    struct kqemu_segment_cache tr;
    struct kqemu_segment_cache gdt; /* only base and limit are used */
    struct kqemu_segment_cache idt; /* only base and limit are used */

    unsigned long cr0;
    unsigned long dummy5;
    unsigned long cr2;
    unsigned long cr3;
    unsigned long cr4;
    uint32_t a20_mask;

    uint64_t efer __attribute__((aligned(8)));

    unsigned long dr0;
    unsigned long dr1;
    unsigned long dr2;
    unsigned long dr3;
    unsigned long dr6;
    unsigned long dr7;

    int cpl; /* currently only 3 */

    uint32_t error_code; /* error_code when exiting with an exception */
    unsigned long next_eip; /* next eip value when exiting with an interrupt */
    unsigned int nb_pages_to_flush; /* number of pages to flush,
                                       KQEMU_FLUSH_ALL means full flush */
#define KQEMU_MAX_PAGES_TO_FLUSH 512
#define KQEMU_FLUSH_ALL (KQEMU_MAX_PAGES_TO_FLUSH + 1)

    long retval;
};

struct kqemu_init {
    uint8_t *ram_base; /* must be page aligned */
    unsigned long ram_size; /* must be multiple of 4 KB */
    uint8_t *ram_dirty; /* must be page aligned */
    uint32_t **phys_to_ram_map; /* must be page aligned */
    unsigned long *pages_to_flush; /* must be page aligned */
};

#define KQEMU_RET_ABORT    (-1)
#define KQEMU_RET_EXCEPTION 0x0000 /* 8 low order bit are the exception */
#define KQEMU_RET_INT       0x0100 /* 8 low order bit are the interrupt */
#define KQEMU_RET_SOFTMMU   0x0200 /* emulation needed (I/O or
                                      unsupported INSN) */
#define KQEMU_RET_INTR      0x0201 /* interrupted by a signal */
#define KQEMU_RET_SYSCALL   0x0300 /* syscall insn */

#ifdef _WIN32
#define KQEMU_EXEC           CTL_CODE(FILE_DEVICE_UNKNOWN, 1, METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS)
#define KQEMU_INIT           CTL_CODE(FILE_DEVICE_UNKNOWN, 2, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#define KQEMU_GET_VERSION    CTL_CODE(FILE_DEVICE_UNKNOWN, 3, METHOD_BUFFERED, FILE_READ_ACCESS)
#else
#define KQEMU_EXEC           _IOWR('q', 1, struct kqemu_cpu_state)
#define KQEMU_INIT           _IOW('q', 2, struct kqemu_init)
#define KQEMU_GET_VERSION    _IOR('q', 3, int)
#endif

#ifdef __KERNEL__
struct kqemu_state;

#define CDECL __attribute__((regparm(0)))

struct kqemu_state * CDECL kqemu_init(struct kqemu_init *d, int max_locked_pages);
struct kqemu_cpu_state * CDECL kqemu_get_cpu_state(struct kqemu_state *s);
long CDECL kqemu_exec(struct kqemu_state *s);
void CDECL kqemu_delete(struct kqemu_state *s);

/* callbacks */
struct kqemu_page; /* opaque data for host page */
struct kqemu_user_page; /* opaque data for host user page */

struct kqemu_user_page *CDECL kqemu_lock_user_page(unsigned long *ppage_index,
                                                   unsigned long user_addr);
void CDECL kqemu_unlock_user_page(struct kqemu_user_page *page);

struct kqemu_page *CDECL kqemu_alloc_zeroed_page(unsigned long *ppage_index);
void CDECL kqemu_free_page(struct kqemu_page *page);
void * CDECL kqemu_page_kaddr(struct kqemu_page *page);

void * CDECL kqemu_vmalloc(unsigned int size);
void CDECL kqemu_vfree(void *ptr);
unsigned long CDECL kqemu_vmalloc_to_phys(const void *vaddr);

void * CDECL kqemu_io_map(unsigned long page_index, unsigned int size);
void CDECL kqemu_io_unmap(void *ptr, unsigned int size);

int CDECL kqemu_schedule(void);

void CDECL kqemu_log(const char *fmt, ...);

#endif

#endif /* KQEMU_H */
